/*
 * UVC library (reference UVC gadget test application)
 *
 * Copyright (C) 2010 Ideas on board SPRL <laurent.pinchart@ideasonboard.com>
 * Copyright 2016 Sony Corporation
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 */
#include <sys/time.h>
#include <sys/ioctl.h>
#include <sys/mman.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/select.h>

#include <unistd.h>
#include <fcntl.h>
#include <stdlib.h>
#include <stdio.h>
#include <stdint.h>
#include <string.h>
#include <errno.h>

#include <linux/usb/ch9.h>
#include <linux/usb/video.h>
#include <linux/videodev2.h>
#include "uvc.h"
#include "libguvc.h"

//#define UVC_DEGBU

#ifdef UVC_DEGBU
#define DBG(x ...)	printf(x)
#else
#define DBG(x ...)
#endif

#define clamp(val, min, max) ({                 \
        typeof(val) __val = (val);              \
        typeof(min) __min = (min);              \
        typeof(max) __max = (max);              \
        (void) (&__val == &__min);              \
        (void) (&__val == &__max);              \
        __val = __val < __min ? __min: __val;   \
        __val > __max ? __max: __val; })

#define ARRAY_SIZE(a)	((sizeof(a) / sizeof(a[0])))

struct uvc_format_info
{
	unsigned int fcc;
	struct uvc_frame_info *frames;
};

struct uvc_device
{
	int fd;
	int mjpeg_fd;

	struct uvc_streaming_control probe;
	struct uvc_streaming_control commit;

	int control;

	unsigned int fcc;
	unsigned int width;
	unsigned int height;
	unsigned int interval;

	void **mem;
	unsigned int nbufs;
	unsigned int bufsize;

	unsigned int imgsize;
	void *imgdata;
	unsigned int now_imgsize;
	char *tmpdata;
	unsigned long rp;
	int isStreaming;
	void *priv;
	struct uvc_ops *ops;
	int wait_data;
	int wait_streaming;

	struct uvc_format_info *format_info;
	int numFormat;
};

static struct uvc_frame_info uvc_frames_yuyv[] = {
	{  640, 360, { 666666, 10000000, 50000000, 0 }, },
	{ 1280, 720, { 50000000, 0 }, },
	{ 0, 0, { 0, }, },
};

static struct uvc_frame_info uvc_frames_mjpeg[] = {
	{  640, 360, { 666666, 10000000, 50000000, 0 }, },
	{ 1280, 720, { 50000000, 0 }, },
	{ 0, 0, { 0, }, },
};

static struct uvc_frame_info uvc_frames_h264[] = {
	{  640, 360, { 666666, 10000000, 50000000, 0 }, },
	{ 0, 0, { 0, }, },
};

static struct uvc_format_info uvc_formats[] = {
	{ V4L2_PIX_FMT_YUYV, uvc_frames_yuyv },
	{ V4L2_PIX_FMT_MJPEG, uvc_frames_mjpeg },
	{ V4L2_PIX_FMT_H264, uvc_frames_h264 },
};

static void
uvc_fill_streaming_control(struct uvc_device *dev,
			   struct uvc_streaming_control *ctrl,
			   int iframe, int iformat)
{
	const struct uvc_format_info *format;
	const struct uvc_frame_info *frame;
	unsigned int nframes;

	if (dev->format_info) {
		if (iformat < 0)
			iformat = dev->numFormat + iformat;
		if (iformat < 0 || iformat >= dev->numFormat)
			return;
		format = &dev->format_info[iformat];	
	} else {
		if (iformat < 0)
			iformat = ARRAY_SIZE(uvc_formats) + iformat;
		if (iformat < 0 || iformat >= (int)ARRAY_SIZE(uvc_formats))
			return;
		format = &uvc_formats[iformat];	
	}

	nframes = 0;
	while (format->frames[nframes].width != 0)
		++nframes;
	if (iframe < 0)
		iframe = nframes + iframe;
	if (iframe < 0 || iframe >= (int)nframes)
		return;
	frame = &format->frames[iframe];
	memset(ctrl, 0, sizeof *ctrl);
	ctrl->bmHint = 1;
	ctrl->bFormatIndex = iformat + 1;
	ctrl->bFrameIndex = iframe + 1;
	ctrl->dwFrameInterval = frame->intervals[0];
	switch (format->fcc) {
	case V4L2_PIX_FMT_YUYV:
		ctrl->dwMaxVideoFrameSize = frame->width * frame->height * 2;
		break;
	case V4L2_PIX_FMT_MJPEG:
		ctrl->dwMaxVideoFrameSize = frame->width * frame->height * 2;
		break;
	}
	ctrl->dwMaxPayloadTransferSize = 512;	/* TODO this should be filled by the driver. */
	ctrl->bmFramingInfo = 3;
	ctrl->bPreferedVersion = 1;
	ctrl->bMaxVersion = 1;
}

static void uvc_events_init(struct uvc_device *dev)
{
	struct v4l2_event_subscription sub;

	uvc_fill_streaming_control(dev, &dev->probe, 0, 0);
	uvc_fill_streaming_control(dev, &dev->commit, 0, 0);

	memset(&sub, 0, sizeof sub);
	sub.type = UVC_EVENT_SETUP;
	ioctl(dev->fd, VIDIOC_SUBSCRIBE_EVENT, &sub);
	sub.type = UVC_EVENT_DATA;
	ioctl(dev->fd, VIDIOC_SUBSCRIBE_EVENT, &sub);
	sub.type = UVC_EVENT_STREAMON;
	ioctl(dev->fd, VIDIOC_SUBSCRIBE_EVENT, &sub);
	sub.type = UVC_EVENT_STREAMOFF;
	ioctl(dev->fd, VIDIOC_SUBSCRIBE_EVENT, &sub);
}

static int uvc_video_init(struct uvc_device *dev __attribute__((__unused__)))
{
	return 0;
}

static void
uvc_video_fill_buffer(struct uvc_device *dev, struct v4l2_buffer *buf)
{
	int fill_size;
	if (dev->ops && dev->ops->fill_buffer) {
		fill_size = dev->ops->fill_buffer(dev, dev->mem[buf->index]);
		if (fill_size > 0)
			buf->bytesused = fill_size;
		else
			buf->bytesused = 0;
	}
}

static int
uvc_video_reqbufs(struct uvc_device *dev, int nbufs)
{
	struct v4l2_requestbuffers rb;
	struct v4l2_buffer buf;
	unsigned int i;
	int ret;

	for (i = 0; i < dev->nbufs; ++i)
		munmap(dev->mem[i], dev->bufsize);

	free(dev->mem);
	dev->mem = 0;
	dev->nbufs = 0;

	memset(&rb, 0, sizeof rb);
	rb.count = nbufs;
	rb.type = V4L2_BUF_TYPE_VIDEO_OUTPUT;
	rb.memory = V4L2_MEMORY_MMAP;

	ret = ioctl(dev->fd, VIDIOC_REQBUFS, &rb);
	if (ret < 0) {
		printf("Unable to allocate buffers: %s (%d).\n",
			strerror(errno), errno);
		return ret;
	}

	DBG("%u buffers allocated.\n", rb.count);

	/* Map the buffers. */
	dev->mem = malloc(rb.count * sizeof dev->mem[0]);

	for (i = 0; i < rb.count; ++i) {
		memset(&buf, 0, sizeof buf);
		buf.index = i;
		buf.type = V4L2_BUF_TYPE_VIDEO_OUTPUT;
		buf.memory = V4L2_MEMORY_MMAP;
		ret = ioctl(dev->fd, VIDIOC_QUERYBUF, &buf);
		if (ret < 0) {
			printf("Unable to query buffer %u: %s (%d).\n", i,
				strerror(errno), errno);
			return -1;
		}
		DBG("length: %u offset: %u\n", buf.length, buf.m.offset);

		dev->mem[i] = mmap(0, buf.length, PROT_READ | PROT_WRITE, MAP_SHARED, dev->fd, buf.m.offset);
		if (dev->mem[i] == MAP_FAILED) {
			printf("Unable to map buffer %u: %s (%d)\n", i,
				strerror(errno), errno);
			return -1;
		}
		DBG("Buffer %u mapped at address %p.\n", i, dev->mem[i]);
	}

	dev->bufsize = buf.length;
	dev->nbufs = rb.count;

	return 0;
}

static int
uvc_video_stream(struct uvc_device *dev, int enable)
{
	struct v4l2_buffer buf;
	unsigned int i;
	int ret = 0;
	int type = V4L2_BUF_TYPE_VIDEO_OUTPUT;

	if (!enable) {
		DBG("Stopping video stream.\n");
		if (dev->ops && dev->ops->stop_stream)
			dev->ops->stop_stream(dev);
		ioctl(dev->fd, VIDIOC_STREAMOFF, &type);
		dev->isStreaming = 0;
		return 0;
	}

	DBG("Starting video stream.\n");
	if (dev->ops && dev->ops->start_stream) {
		if (dev->ops->start_stream(dev, dev->fcc, dev->width,
								   dev->height, dev->interval) < 0)
			return 0;
	}
	for (i = 0; i < dev->nbufs; ++i) {
		memset(&buf, 0, sizeof buf);

		buf.index = i;
		buf.type = V4L2_BUF_TYPE_VIDEO_OUTPUT;
		buf.memory = V4L2_MEMORY_MMAP;

		uvc_video_fill_buffer(dev, &buf);

		DBG("Queueing buffer %u.\n", i);
		if ((ret = ioctl(dev->fd, VIDIOC_QBUF, &buf)) < 0) {
			printf("Unable to queue buffer: %s (%d).\n",
				strerror(errno), errno);
			break;
		}
	}
	ioctl(dev->fd, VIDIOC_STREAMON, &type);
	dev->isStreaming = 1;
	return ret;
}

static int
uvc_video_set_format(struct uvc_device *dev)
{
	struct v4l2_format fmt;
	int ret;

	DBG("Setting format to 0x%08x %ux%u\n",
		dev->fcc, dev->width, dev->height);

	memset(&fmt, 0, sizeof fmt);
	fmt.type = V4L2_BUF_TYPE_VIDEO_OUTPUT;
	fmt.fmt.pix.width = dev->width;
	fmt.fmt.pix.height = dev->height;
	fmt.fmt.pix.pixelformat = dev->fcc;
	fmt.fmt.pix.field = V4L2_FIELD_NONE;
	if (dev->fcc == V4L2_PIX_FMT_MJPEG || V4L2_PIX_FMT_H264)
		fmt.fmt.pix.sizeimage = dev->width * dev->height  * 2;

	if ((ret = ioctl(dev->fd, VIDIOC_S_FMT, &fmt)) < 0)
		printf("Unable to set format: %s (%d).\n",
			strerror(errno), errno);

	return ret;
}

static void
uvc_events_process_standard(struct uvc_device *dev, struct usb_ctrlrequest *ctrl,
			    struct uvc_request_data *resp)
{
	DBG("standard request\n");
	(void)dev;
	(void)ctrl;
	(void)resp;
}

static void
uvc_events_process_streaming(struct uvc_device *dev, uint8_t req, uint8_t cs,
			     struct uvc_request_data *resp)
{
	struct uvc_streaming_control *ctrl;

	DBG("streaming request (req %02x cs %02x)\n", req, cs);

	if (cs != UVC_VS_PROBE_CONTROL && cs != UVC_VS_COMMIT_CONTROL)
		return;

	ctrl = (struct uvc_streaming_control *)&resp->data;
	resp->length = sizeof *ctrl;

	switch (req) {
	case UVC_SET_CUR:
		dev->control = cs;
		resp->length = 0;
		dev->wait_data = 1;
		break;

	case UVC_GET_CUR:
		if (cs == UVC_VS_PROBE_CONTROL)
			memcpy(ctrl, &dev->probe, sizeof *ctrl);
		else
			memcpy(ctrl, &dev->commit, sizeof *ctrl);
		break;

	case UVC_GET_MIN:
	case UVC_GET_MAX:
	case UVC_GET_DEF:
		uvc_fill_streaming_control(dev, ctrl, req == UVC_GET_MAX ? -1 : 0,
					   req == UVC_GET_MAX ? -1 : 0);
		break;

	case UVC_GET_RES:
		memset(ctrl, 0, sizeof *ctrl);
		break;

	case UVC_GET_LEN:
		resp->data[0] = 0x00;
		resp->data[1] = 0x22;
		resp->length = 2;
		break;

	case UVC_GET_INFO:
		resp->data[0] = 0x03;
		resp->length = 1;
		break;
	}
}

static void
uvc_events_process_class(struct uvc_device *dev, struct usb_ctrlrequest *ctrl,
			 struct uvc_request_data *resp)
{
	if ((ctrl->bRequestType & USB_RECIP_MASK) != USB_RECIP_INTERFACE)
		return;

	switch (ctrl->wIndex & 0xff) {
	case UVC_INTF_CONTROL:
		uvc_events_process_streaming(dev, ctrl->bRequest, ctrl->wValue >> 8, resp);
		break;

	case UVC_INTF_STREAMING:
		uvc_events_process_streaming(dev, ctrl->bRequest, ctrl->wValue >> 8, resp);
		break;

	default:
		break;
	}
}

static void
uvc_events_process_setup(struct uvc_device *dev, struct usb_ctrlrequest *ctrl,
			 struct uvc_request_data *resp)
{
	DBG("bRequestType %02x bRequest %02x wValue %04x wIndex %04x "
		"wLength %04x\n", ctrl->bRequestType, ctrl->bRequest,
		ctrl->wValue, ctrl->wIndex, ctrl->wLength);

	switch (ctrl->bRequestType & USB_TYPE_MASK) {
	case USB_TYPE_STANDARD:
		uvc_events_process_standard(dev, ctrl, resp);
		break;

	case USB_TYPE_CLASS:
		uvc_events_process_class(dev, ctrl, resp);
		break;

	default:
		break;
	}
}

static void
uvc_events_process_data(struct uvc_device *dev, struct uvc_request_data *data)
{
	struct uvc_streaming_control *target;
	struct uvc_streaming_control *ctrl;
	const struct uvc_format_info *format;
	const struct uvc_frame_info *frame;
	const unsigned int *interval;
	unsigned int iformat, iframe;
	unsigned int nframes;

	switch (dev->control) {
	case UVC_VS_PROBE_CONTROL:
		DBG("setting probe control, length = %d\n", data->length);
		target = &dev->probe;
		break;

	case UVC_VS_COMMIT_CONTROL:
		DBG("setting commit control, length = %d\n", data->length);
		target = &dev->commit;
		break;

	default:
		DBG("setting unknown control, length = %d\n", data->length);
		return;
	}

	ctrl = (struct uvc_streaming_control *)&data->data;
	if (dev->format_info) {
		iformat = clamp((unsigned int)ctrl->bFormatIndex, 1U,
						(unsigned int)dev->numFormat);
		format = &dev->format_info[iformat-1];
	} else {
		iformat = clamp((unsigned int)ctrl->bFormatIndex, 1U,
						(unsigned int)ARRAY_SIZE(uvc_formats));
		format = &uvc_formats[iformat-1];
	}

	nframes = 0;
	while (format->frames[nframes].width != 0)
		++nframes;

	iframe = clamp((unsigned int)ctrl->bFrameIndex, 1U, nframes);
	frame = &format->frames[iframe-1];
	interval = frame->intervals;

	while (interval[0] < ctrl->dwFrameInterval && interval[1])
		++interval;

	target->bFormatIndex = iformat;
	target->bFrameIndex = iframe;
	switch (format->fcc) {
	case V4L2_PIX_FMT_YUYV:
		target->dwMaxVideoFrameSize = frame->width * frame->height * 2;
		break;
	case V4L2_PIX_FMT_MJPEG:
		target->dwMaxVideoFrameSize = frame->width * frame->height * 2;
		break;
	}
	target->dwFrameInterval = *interval;

	if (dev->control == UVC_VS_COMMIT_CONTROL) {
		dev->fcc = format->fcc;
		dev->width = frame->width;
		dev->height = frame->height;
		dev->interval = *interval;

		uvc_video_set_format(dev);
	}
}

static int
uvc_video_process(struct uvc_device *dev)
{
	struct v4l2_buffer buf;
	int ret;

	if (!dev->isStreaming)
		return 0;

	memset(&buf, 0, sizeof buf);
	buf.type = V4L2_BUF_TYPE_VIDEO_OUTPUT;
	buf.memory = V4L2_MEMORY_MMAP;

	if ((ret = ioctl(dev->fd, VIDIOC_DQBUF, &buf)) < 0) {
		printf("Unable to dequeue buffer: %s (%d).\n", strerror(errno),
			   errno);
		return ret;
	}

	uvc_video_fill_buffer(dev, &buf);

	if ((ret = ioctl(dev->fd, VIDIOC_QBUF, &buf)) < 0) {
		printf("Unable to requeue buffer: %s (%d).\n", strerror(errno),
			errno);
		return ret;
	}

	return 0;
}

static void
uvc_events_process(struct uvc_device *dev)
{
	struct v4l2_event v4l2_event;
	struct uvc_event *uvc_event = (void *)&v4l2_event.u.data;
	struct uvc_request_data resp;
	int ret;

	ret = ioctl(dev->fd, VIDIOC_DQEVENT, &v4l2_event);
	if (ret < 0) {
		printf("VIDIOC_DQEVENT failed: %s (%d)\n", strerror(errno),
			errno);
		return;
	}

	memset(&resp, 0, sizeof resp);
	resp.length = 0;

	switch (v4l2_event.type) {
	case UVC_EVENT_CONNECT:
	case UVC_EVENT_DISCONNECT:
		return;

	case UVC_EVENT_SETUP:
		uvc_events_process_setup(dev, &uvc_event->req, &resp);
		ioctl(dev->fd, UVCIOC_SEND_RESPONSE, &resp);
		if (ret < 0) {
			printf("UVCIOC_S_EVENT failed: %s (%d)\n", strerror(errno),
				   errno);
			return;
		}
		break;

	case UVC_EVENT_DATA:
		dev->wait_data = 0;
		uvc_events_process_data(dev, &uvc_event->data);
		if (dev->wait_streaming) {
			dev->wait_streaming = 0;
			uvc_video_reqbufs(dev, 4);
			uvc_video_stream(dev, 1);
		}
		return;

	case UVC_EVENT_STREAMON:
		if (dev->wait_data) {
			dev->wait_streaming = 1;
		} else {
			uvc_video_reqbufs(dev, 4);
			uvc_video_stream(dev, 1);
		}
		resp.length=34;
		break;

	case UVC_EVENT_STREAMOFF:
		dev->wait_streaming = 0;
		uvc_video_stream(dev, 0);
		uvc_video_reqbufs(dev, 0);
		resp.length=34;
		break;
	}
}

static int create_uvc_format(struct uvc_device *dev)
{
	struct uvc_format_desc format_desc;
	struct uvc_frame_desc frame_desc;
	int i, j;
	int numFormat;
	
	if (ioctl(dev->fd, UVCIOC_GET_NUMFORMAT, &numFormat) < 0)
		return -1;

	if (numFormat <= 0)
		return -1;

	dev->numFormat = numFormat;
	dev->format_info = calloc(numFormat, sizeof(struct uvc_format_info));
	if (!dev->format_info) {
		printf("format info memory allcate failed.\n");
		return -1;
	}

	for (i=0; i<numFormat; i++) {
		format_desc.bFormatIndex = i + 1;
		if (ioctl(dev->fd, UVCIOC_GET_FORMATDESC, &format_desc) < 0)
			goto Err;
		if (format_desc.bNumFrameDescriptors <= 0)
			continue;
		dev->format_info[i].fcc = format_desc.fcc;
		dev->format_info[i].frames = 
			calloc(format_desc.bNumFrameDescriptors + 1,
				   sizeof(struct uvc_frame_info));
		if (!dev->format_info[i].frames) {
			printf("frame info memory allcate failed.\n");
			goto Err;
		}

		for (j=0; j < format_desc.bNumFrameDescriptors; j++) {
			frame_desc.bFormatIndex = i + 1;
			frame_desc.bFrameIndex = j + 1;
			if (ioctl(dev->fd, UVCIOC_GET_FRAMEDESC, &frame_desc) < 0)
				goto Err;
			memcpy((void *)&dev->format_info[i].frames[j],
				   &frame_desc.frame_info, sizeof(struct uvc_frame_info));
		}
	}

	return 0;

 Err:
	for (i=0; i<dev->numFormat; i++) {
		if (dev->format_info[i].frames)
			free(dev->format_info[i].frames);
	}
	
	free(dev->format_info);
	dev->format_info = NULL;

	return -1;
}

struct uvc_device *uvc_open(const char *devname)
{
	struct uvc_device *dev;
	int fd;

	/* dummy open */
	fd = open(devname, O_RDWR | O_NONBLOCK);
	if (fd >= 0) {
		close(fd);
	}

	fd = open(devname, O_RDWR | O_NONBLOCK);
	if (fd == -1) {
		printf("v4l2 open failed: %s (%d)\n", strerror(errno), errno);
		return NULL;
	}

	DBG("open succeeded, file descriptor = %d\n", fd);

	dev = malloc(sizeof *dev);
	if (dev == NULL) {
		close(fd);
		return NULL;
	}

	memset(dev, 0, sizeof *dev);
	dev->fd = fd;

	create_uvc_format(dev);

	uvc_events_init(dev);
	uvc_video_init(dev);

	return dev;
}

void uvc_close(struct uvc_device *dev)
{
	int i;
	close(dev->fd);
	free(dev->imgdata);
	free(dev->mem);
	if (dev->format_info) {
		for (i=0; i<dev->numFormat; i++) {
			if (dev->format_info[i].frames)
				free(dev->format_info[i].frames);
		}
		free(dev->format_info);
		dev->format_info = NULL;
		
	}
	free(dev);
}

int uvc_register_ops(struct uvc_device *dev, struct uvc_ops *ops)
{
	dev->ops = ops;

	return 0;
}

int uvc_start(struct uvc_device *dev)
{
	fd_set fds;
	int ret;

	FD_ZERO(&fds);
	FD_SET(dev->fd, &fds);

	while (1) {
		fd_set efds = fds;
		fd_set wfds = fds;

		ret = select(dev->fd + 1, NULL, &wfds, &efds, NULL);
		if (ret < 0)
			break;

		if (FD_ISSET(dev->fd, &efds))
			uvc_events_process(dev);
		if (FD_ISSET(dev->fd, &wfds)) {
			uvc_video_process(dev);
		}
	}

	return 0;
}

void uvc_register_private_data(struct uvc_device *dev, void *data)
{
	if (dev)
		dev->priv = data;
}

void *uvc_device_to_priv(struct uvc_device *dev)
{
	if (!dev)
		return NULL;

	return dev->priv;
}

